
/* Functions using MFRAG and PAIRCOMP */

/* Parts of source file (line numbers approximate): */

/* (***AAA) (L29) Fragment collection routines:  */
/*             do_inner(), do_outerseq(), etc. */

/* (***BBB) (L567) Init_array() and friends */

/* (***CCC) (L754) General utilities: */
/*            (including Karlin-Altschul routines) */


#include "vall.h"
#include "vflags.h"
#include "vread.h"
#include "vcalc.h"

SIMX cfinx, cfoutseqx, cfoutpairx, cfoutgrpx;

SIMTYPE runsimtype=SIM_NONE;

char *AllRightsReserved_Vcalc =
   "All rights reserved - Stanley Sawyer";


/* (***AAA)  Fragment collection routines:  */
/*             For do_inner(), do_outerseq(), etc. */

MFRAG *permlist, *currlist;
int npmkmax,npkmax,listpkeep, ncurr,ncurrmax,nperm, scoregate;
#ifdef VTEXT
int setupflag, adsimdone;
#else /* Essential in VSHOW */
int setupflag=TRUE;
#endif
int simscore,simcurr;
int circ2flag, maxfragstart, vnumpairs;

static int usekapw, usekaglob, useglobal_for_pairw;
int havbarrier, havpwbarrier, havglob_barrier;
double kabarrier, kapwbarrier, kaglob_barrier;
int tothavcirc;

SRUN neg, pos;   /* For runs of negative,positive site scores */

static void getlakmax (PAIRCOMP *pc);
static void circback (int i, int j, int first, int last);
static int sortstart (const void *p, const void *q);


/* Fragment collection arrays and functions       */
/* Most of the overhead is for collecting the     */
/*   `listpkeep' best fragments for each pair     */
/*   If VTEXT, the `listglobal' best fragments are   */
/*   collected in vtcalc.c                        */
/* Here listpkeep = max(listperpair[listpairs],listglobal)        */
/*   if VTEXT (otherwise listpkeep=listperpair)       */
/* addfrag() is called only if setupflag==TRUE or if VSHOW    */

/* addsim() finds the maximum pairwise score in each simulation */
/*   for permutation-test P-values */

/* addfrag() constructs two lists: The  ncurr  highest-scoring fragments */
/*   ending at the current position and the  nperm  highest-scoring */
/*   fragments so far, without regards to endpoints.  A similar */
/*   list is collected in vtcalc.c for globally significant fragments. */
/* addsim() (in VTEXT only) keeps track of three scores: */
/*   sumcurr, the highest score for a fragment ending at the current */
/*   position, simscore, the highest score so far, and globscore, */
/*   the global analog. */

/* addfrag() and addsim() are called at the end of runs of matches, */
/*   which are generally preceded by a run of mismatches.  Information */
/*   about the matches is in (SRUN)pos and about the mismatches is */
/*   in (SRUN)neg.  The information in (SRUN)neg is not used */
/*   on the first call to addfrag or addsim */

/* Both functions proceed in three steps: */
/*    (i) use (pos) or (neg,pos) to update currlist[] (or simcurr), */
/*   (ii) compare or fit in (pos) with currlist[] or simcurr, */
/*  (iii) use currlist[] (or simcurr) to update permlist[] */
/*             (or simscore) */

/* Fragments are stored as [start,fin) in polymorphism # s, where */
/*   start<fin.  Thus fin=start+npolys means the entire alignment, */
/*   but this is excluded sinces its score <=0  and minscore>=1 */
/*   It seems likely that fragments with fin-start=npolys could be */
/*   excluded. */

/* addfrag() and addsim() are entered only if pos.valid==TRUE. */
/*   Then  pos.start is the first MATCHING polymorphism # in a run */
/*     (That is, the run is [a,b) = [pos.start,rfin) in poly # s) */
/*   pos.score is the sum of scores for the run of matches */
/*   rfin-pos.start is the number of polymorphisms in the run, */
/*   even when pos.start<0 or rfin>npolys for ``round-the-corner' */
/*   runs when circular==TRUE.
/* neg.score is the sum of scores for the PRECEDING run of mismatches */
/*   neg.score and neg.npoly refer to the run of mismatches */
/*   neg.start and pos.npoly are not used. */
/* The number of polymorphisms in a fragment */
/*   is always (rfin-pos.start) */


/* NOTES ABOUT ``ROUND-THE-CORNER'' FRAGMENTS and RUNS: */

/* A ``round-the-corner'' fragment or run can arise ONLY */
/*   in one of the two following situations (if circular==TRUE): */
/*     (i) The initial run is ``stitched'' back to a polymorphism */
/*   number start<0.  In that case, `maxfragstart' is decreased */
/*   to help avoid fragments that more than encircle the genome. */
/*     (ii) The run ends in the second loop through polysites[]  */
/*   with rfin > npolys.  In that case circ2flag==TRUE, and the ending */
/*   polymorphism number is in the range  npolys < rfin < 2*npolys.  */

/*     Since the conditions ``fin-sta<=npolys'' is imposed in */
/*   addfrag(r), a surviving fragment one cannot have runs of types */
/*   BOTH (i) and (ii).  */

/*     The code in TOP_STUFF guarantees that there is at least one */
/*   mismatch in the alignment.  Otherwise, the initial ``backstitched'' */
/*   run would be the entire genome.  */

/*     Thus if circular==TRUE, three types of fragments are saved: */
/*       (i) 0<=start<finish<=npolys, which are converted to (fbase,lbase) */
/*   with  1<=fbase<=lbase<=nbases.  */
/*      (ii) start<0<finish<npolys, which are converted to (fbase,lbase) */
/*   with 1<=lbase<fbase<=nbases.  */
/*     (iii) 0<=start<npolys<finish, which are converted as in (ii) */
/*                                        */
/*   In all cases,  caflag = (lbase<fbase);    */
/*                                        */


void addfrag(int rfin)
 { int j;  MFRAG fr;
   if (circ2flag)  rfin+=npolys;

   /* Step I: Use (pos) on the first call (ncurr==0) */
   /*   Use (neg,pos) on succeeding calls   to update currlist[] */
   /* If gscale==0, then the mismatch penalty is infinite */
   /*   and currlist[] will not be used beyond currlist[0] */

   if (gscale==0)  ncurr=0;
   else if (ncurr>0)
    { int sumscore;
      /* If circ2flag==TRUE, then new entries will not be added */
      /*   to currlist[], and fragments that now encircle the genome */
      /*   should be dropped.  Currlist[] is sorted by `start' */
      /*   when circ2flag is set =TRUE. */
      if (circ2flag)
        while (ncurr>0 && rfin-currlist[ncurr-1].start>npolys)
          ncurr--;
      /* Use (neg) and (pos) to update (prior) currlist[] */
      sumscore=neg.score+pos.score;
      for (j=0; j<ncurr; j++)
       { currlist[j].score+=sumscore;
         currlist[j].finish=rfin;
         currlist[j].nmismat+=neg.npoly; }}

   /* Step II: See if (pos) by itself qualifies for currlist[] */
   /* The (listperpair+1)-th entry is needed to record last-place ties */
   /* Generally circ2flag==TRUE iff rfin >= npolys, although */
   /*   a final addfrag(npolys) call is made to flush the queue */
   /*   if circular==FALSE and pos.valid==TRUE.  */
   /* Skip Step II if we have seen this run before */
   /*   (for example, if pos.start>=npolys) */
   if (circ2flag && pos.start>=maxfragstart)  goto dont_add;
   if (ncurr>=npkmax)  /* Is currlist[] full? */
     currlist=rexxalloc(currlist,sizeof(MFRAG),
        ncurr, npkmax+=FRAGBAT);
   /* Add (pos) by itself to currlist[]: */
   memset(&fr,0,sizeof(MFRAG));
   fr.start=pos.start;  fr.finish=rfin;  fr.score=pos.score;
   currlist[ncurr++]=fr;  ncurrmax=ncurr;

dont_add:
   /* Step III: Update permlist[] from currlist[]. */
   /* The (listperpair+1)-th entry is needed to record last-place ties */
   /* Currlist[] is NOT sorted by score */
   for (j=0; j<ncurr; j++)
     if (currlist[j].score>=minscore)
      { int fin=currlist[j].finish, sta=currlist[j].start;
        if (fin-sta > npolys || fin-sta < minnpoly
           || (havbarrier && currlist[j].score < kabarrier)
           || (minlength>1 && offsetlen(sta,fin)<minlength))
                continue;
        if (nperm>=npmkmax)  /* Is permlist[] full? */
          permlist=rexxalloc(permlist,sizeof(MFRAG),
             nperm, npmkmax+=FRAGBAT);
        /* Add an MFRAG to permlist[]: */
        permlist[nperm++]=currlist[j]; }}
   /* End of function  addfrag(r) */



static double log1mp (double pval)
 { return (pval<1e-8) ? pval : -log(1.0-pval); }


static void setup_consts (PAIRCOMP *pc)
 { int totdifs=pc->totdifs;
   /* Set the mismatch penalty at -gscale * npolys/totdifs */
   /* Divide gscale*npolys by totdifs, rounding up */
   if (gscale>0 && totdifs>0)
      pc->mismatpen = -((gscale*npolys + totdifs-1)/totdifs);
   /* If (gscale==0 || totdifs==0), mismatches cannot occur */
   else  pc->mismatpen=-800;
   /* Define  pc->ll, pc->kmax, and pc->kavalid */
   /*   for Karlin-Altschul approximate P-values */
   getlakmax(pc);

   /* Can we apply an apriori lower barrier on fragment scores */
   /*   based on KA Pvalues? */
   /* If VTEXT is defined, then there are both pairwise and global lists. */
   /*   All a prior barriers apply to BOTH lists. Further restrictions */
   /*   based on KA Pvalues are applied in `set_pfrags()' in vtcalc.c.  */
   /*   If there are conditions on both maxkapairpval and maxkaglobpval, */
   /*   use the smaller of the two corresponding lower barriers. */
   /*   If there is a condition on only one of the two, use that */
   /*   barrier for both lists (to avoid very long lists). */
   /* If bcsims==TRUE, then the PAIRCOMP score-level arrays */
   /*   will have to be ample enough for global fragments for that pair. */
   /* If VSHOW is defined, then there are only pairwise lists. */
   /*   Thus the lower barrier can be based on the smaller or only */
   /*   KA P-value (or the larger barrier).  There will be no further */
   /*   restrictions based on KA Pvalues. */
   havpwbarrier=havglob_barrier=FALSE;
   if (pc->kavalid)
    { havpwbarrier=usekapw;  havglob_barrier=usekaglob;
      havbarrier = (havpwbarrier || havglob_barrier);
      /* Havbarrier and kabarrier apply to PAIRCOMP lists */
      /* A fragment score must be >= `kabarrier' to be accepted. */
      /* `Useglobal_for_pairw' is set in two cases: */
      /*    (i) No pairwise KA maximum P-value is given */
      /*    (ii) A pairwise KA P-value is given, but it is too */
      /*      restrictive to feed all fragments to global lists */
      if (havbarrier)
       { double km=pc->kmax, ll=pc->ll;
         if (havpwbarrier)
           kapwbarrier=ceil(log(km*npolys/log1mp(maxkapairpval))/ll);
         if (havglob_barrier)
           kaglob_barrier=ceil(log(km*npolys
               /log1mp(maxkaglobpval/vnumpairs))/ll);
         kabarrier=(useglobal_for_pairw) ? kaglob_barrier : kapwbarrier; }}

   /* Low-level information if -Test=<Pairlist> is entered. */
   if (match_showparms(pc->seq1,pc->seq2))
    { char q1b[MAXNAMETWO];
      BPRINTF ("BLAST-like minimum scores (%s):\n",
        fragcseqs(q1b,pc,runsimtype));
      if (pc->totdifs==npolys || pc->totdifs==0)
        BPRINTF (
        "  Parameters are invalid and there are no fragments.\n");
      else if (pc->kavalid==FALSE)
        BPRINTF ("  Parameters are invalid (no barriers).\n");
      else if (havbarrier)
       { char pbuff[33]="No limit", gbuff[33]="No limit";
         if (havpwbarrier) sprintf(pbuff, "%g",kapwbarrier);
         if (havglob_barrier) sprintf(pbuff, "%g",kaglob_barrier);
         BPRINTF("  Pairwise: %s   Global: %s\n", pbuff,gbuff); }
      else BPRINTF ("  No barriers.\n"); }

   /* Initialize local fragment collection arrays */
   ncurr=ncurrmax=nperm=0;
   memset(currlist, 0, npkmax*sizeof(currlist[0]));
   memset(permlist, 0, npmkmax*sizeof(permlist[0]));
       }


/* Return if there are no possible fragments or one fragment */
/* If pc->totdifs==0, then pc->kavalid==FALSE */

#ifdef VTEXT

#define TOP_STUFF   if (pc->totdifs==npolys || pc->totdifs==0)  return;   \
                    simscore=simcurr=0;                              \
                    pos.valid=circ2flag=adsimdone=FALSE;             \
                    mismatpen=pc->mismatpen;                         \
                    maxfragstart=npolys

#define END_STUFF   if (setupflag)  set_pfrags(pc);        \
                    else  update_pscores(pc)


#else /* if VSHOW */

#define TOP_STUFF   if (pc->totdifs==0 || pc->totdifs==npolys      \
                          || pc->kavalid==FALSE)  return;      \
                    simscore=simcurr=0;                        \
                    pos.valid=circ2flag=FALSE;                 \
                    mismatpen=pc->mismatpen;                   \
                    maxfragstart=npolys

#define END_STUFF   set_pfrags(pc)

#endif /* VTEXT */


/* INNER_LOOP and friends: */

/* If VSHOW or setupflag==TRUE, find and sort appropriate fragments */
/*    and call  set_pfrags()             */
/* If VTEXT and setupflag==FALSE, compute simscore */
/*    and call  update_pscores(pc)  */

/* Match intervals are described as [a,b), where a,b are */
/*   polymorphism numbers (i.e., offsets in polysite[]). */
/* Pos.npoly and neg.start are not used */

#ifdef VTEXT
void addsim(int r);

#define ADD_SEG(r)  if (setupflag) addfrag(r);      \
                    addsim(r)

#define CIRC_CHECK   if (setupflag) { if (ncurr==0) break; }  \
                     else if (adsimdone)  break

#else /* if VSHOW */

#define ADD_SEG(r)  addfrag(r)

#define CIRC_CHECK  if (ncurr==0) break

#endif /* VTEXT */


#define INNER_LOOP(sc,r)  if (sc >= 0)   /* A match */          \
    { if (pos.valid==FALSE)   /* Initiate a run of matches */   \
       { pos.start=r;  pos.score=0;                       \
         if (circ2flag)  pos.start+=npolys;               \
         pos.valid=TRUE; }                                \
      /* Continue a run of matches */                     \
      pos.score+=sc; }                                    \
   else                  /* A mismatch */                 \
    { if (pos.valid==TRUE)    /* Register a run */        \
       { ADD_SEG(r);      /* Add a run */                 \
         /* Initiate a run of mismatches */               \
         pos.valid=FALSE;  neg.score=neg.npoly=0;         \
         if (circ2flag)                                   \
          { CIRC_CHECK; }}                                \
      /* Continue a run of mismatches */                  \
      /* (Values not used in an initial mismatch run) */  \
      neg.npoly++;  neg.score+=sc; }


/* Routines  do_inner(), do_outerseq(), etc. and friends */


void do_inner (PAIRCOMP *pc)
 { POLYSITE *rr;
   int r, i1=pc->seq1, i2=pc->seq2, mismatpen;
   if (setupflag)  /* First run for this PAIRCOMP */
    { int totdifs=0;
      /* Find the number of mismatched sites between strains #i and #j: */
      for (rr=polysites; rr<sitend; rr++)
        if (rr->site[i1]!=rr->site[i2])  totdifs++;
      pc->totdifs=totdifs;
      setup_consts(pc); }
   /* Return if no fragments and initialize variables */
   /*   for the polymorphism loop */
   TOP_STUFF;

   /* Loop over polymorphic sites: */
   /* Keep track of runs as [a,b), where a,b represent polymorphism #s */
   /* Do not start a run at a polymorphism # >= maxfragstart */
   /* Circular back for first fragment if nec. */
   if (circular && setupflag)  circback(i1,i2,0,0);
do_circular:   /* Loop around again, if circular */
   /* Look for matching bases in strains #i and #j: */
   for (r=0, rr=polysites;  rr<sitend;  rr++, r++)
    { int sc = (rr->site[i1]==rr->site[i2]) ? 1 : mismatpen;
      INNER_LOOP(sc,r); }
   /* End of polymorphism loop */
   if (circular && circ2flag==FALSE)
     { circ2flag=TRUE;
       /* There will be new additions to currlist, so sort on start */
       if (setupflag && ncurr>0)
          qsort(currlist,ncurr,sizeof(currlist[0]),sortstart);
       goto do_circular; }

   END_STUFF; }




void do_outerpair (PAIRCOMP *pc)
 { POLYSITE *rr;
   int r, i1=pc->seq1, i2=pc->seq2, mismatpen;
   if (setupflag)  /* First run for this PAIRCOMP */
    { int totdifs=0;
      /* Number of MATCHING sites in strains #i and #j: */
      for (rr=polysites; rr<sitend; rr++)
        if (rr->site[i1]==rr->site[i2])  totdifs++;
      pc->totdifs=totdifs;  setup_consts(pc); }
   /* Return if no fragments and initialize variables */
   /*   for the polymorphism loop */
   TOP_STUFF;

   /* Loop over polymorphic sites: */
   /* Keep track of runs as [a,b), where a,b represent polymorphism #s */
   /* Do not start a run at a polymorphism # >= maxfragstart */
   /* Circular back for first fragment if nec. */
   if (circular && setupflag)  circback(i1,i2,-1,0);
do_circular:   /* Loop around again, if circular */
   /* Look for non-matching bases in strains #i and #j: */
   for (r=0, rr=polysites;  rr<sitend;  rr++, r++)
    { int sc = (rr->site[i1]!=rr->site[i2]) ? 1 : mismatpen;
      INNER_LOOP(sc,r); }
   /* End of polymorphism loop */
   if (circular && circ2flag==FALSE)
     { circ2flag=TRUE;
       /* There will be new additions to currlist, so sort on start */
       if (setupflag && ncurr>0)
          qsort(currlist,ncurr,sizeof(currlist[0]),sortstart);
       goto do_circular; }

   END_STUFF; }




void do_outerseq (PAIRCOMP *pc)
 { POLYSITE *rr;
   int r, i=pc->seq1, mismatpen;
   int first=grouplist[pc->group].first;
   int last=grouplist[pc->group].last;
   if (setupflag)
    { int j,totdifs=0;
      /* Number of sites with nonunique bases in that group: */
      for (rr=polysites; rr<sitend; rr++)
       { char base, *site=rr->site;  base=site[i];
         for (j=first; j<last; j++)
           if (j!=i && site[j]==base)  { totdifs++;  break; }}
      pc->totdifs=totdifs;  setup_consts(pc); }
   /* Return if no fragments and initialize variables */
   /*   for the polymorphism loop */
   TOP_STUFF;

   /* Loop over polymorphic sites */
   /* Keep track of runs as [a,b), where a,b represent polymorphism # */
   /* Do not start a run at polymorphism # >= maxfragstart */
   /* Circular back for first fragment if nec. */
   if (circular && setupflag)  circback(i,-1,first,last);
do_circular:   /* Loop around again, if circular */
   /* Look for unique bases in that group: */
   for (r=0, rr=polysites;  rr<sitend;  rr++, r++)
    { char  base, *site = rr->site;  int sc,j;
      base=site[i];  sc=1;
      for (j=first; j<last; j++)
         if (site[j]==base && j!=i) { sc=mismatpen;  break; }
      INNER_LOOP(sc,r); }
   /* End of polymorphism loop */
   if (circular && circ2flag==FALSE)
     { circ2flag=TRUE;
       /* No new additions to currlist, so sort on start */
       if (setupflag && ncurr>0)
          qsort(currlist,ncurr,sizeof(currlist[0]),sortstart);
       goto do_circular; }

   END_STUFF; }



void do_outergrp (PAIRCOMP *pc)
 { POLYSITE *rr;
   int r, mismatpen;
   int first=grouplist[pc->group].first;
   int last=grouplist[pc->group].last;
   if (setupflag)
    { int j,totdifs=0;
      /* The number of sites monomorphic in that group: */
      for (rr=polysites; rr<sitend; rr++)
       { char base, *site=rr->site;  base=site[first];
         for (j=first+1; j<last; j++)  if (site[j]!=base)  break;
         if (j>=last)  totdifs++; }
      pc->totdifs=totdifs;  setup_consts(pc); }
   /* Return if no fragments and initialize variables */
   /*   for the polymorphism loop */
   TOP_STUFF;

   /* Loop over polymorphic sites */
   /* Keep track of runs as [a,b), where a,b represent polymorphism # */
   /* Do not start a run at polymorphism # >= maxfragstart */
   /* Circular back for first fragment if nec. */
   if (circular && setupflag)  circback(-1,-1,first,last);
do_circular:   /* Loop around again, if circular */
   /* Look for sites that are polymorphic in that group: */
   for (r=0, rr=polysites;  rr<sitend;  rr++, r++)
    { char base, *site = rr->site;  int sc,j;
      base=site[first];  sc=mismatpen;
      for (j=first; j<last; j++)
        if (site[j]!=base) { sc=1;  break; }
      INNER_LOOP(sc,r); }
   /* End of polymorphism loop */
   if (circular && circ2flag==FALSE)
     { circ2flag=TRUE;
       /* No new additions to currlist, so sort on start */
       if (setupflag && ncurr>0)
          qsort(currlist,ncurr,sizeof(currlist[0]),sortstart);
       goto do_circular; }

   END_STUFF; }




/* Called if circular==TRUE */
/* Back up a maximal segment of matching bases if possible */

static void circback (int i, int j, int first, int last)
 { int r,sc;
   if (j>=0)  /* Pairwise fragments */
    { if (first>=0)  /* Inner pair */
       { /* Do the two initial bases match? */
         if (polysites[0].site[i]==polysites[0].site[j])
          { r=npolys;  sc=0;
            while (--r>0 && polysites[r].site[i]==polysites[r].site[j])
               sc+=1;
            goto start_back; }}
      else   /* Outer pair */
       { /* Do the two initial bases mismatch? */
         if (polysites[0].site[i]!=polysites[0].site[j])
          { r=npolys;  sc=0;
            while (--r>0 && polysites[r].site[i]!=polysites[r].site[j])
               sc+=1;
            goto start_back; }}}
   else if (i>=0)  /* Regular outer fragments */
    { /* Is the initial base unique in that group? */
      char base=polysites[0].site[i];
      for (j=first; j<last; j++)
        if (j!=i && polysites[0].site[j]==base)  break;
      if (j>=last)
       { r=npolys;  sc=0;
         while (--r>0)
          { char base, *site=polysites[r].site;  base=site[i];
            for (j=first; j<last; j++)
              if (j!=i && site[j]==base)  break;
            if (j>=last)  sc+=1;  else break; }
         goto start_back; }}
   else       /* Alternative outer fragments */
    { /* Is the initial base polymorphic in that group? */
      char base=polysites[0].site[first];
      for (j=first+1; j<last; j++)
        if (polysites[0].site[j]!=base)  break;
      if (j<last)
       { r=npolys;  sc=0;
         while (--r>0)
          { char base, *site=polysites[r].site;  base=site[first];
            for (j=first+1; j<last; j++)
              if (site[j]!=base)  break;
            if (j<last)  sc+=1;  else break; }
         goto start_back; }}
   return;  /* Initial mismatch, circback not needed */
start_back:
   pos.valid=TRUE;  pos.start=r+1-npolys;
   pos.score=sc;  maxfragstart=r+1; }



int sortpfrag (const void *p, const void *q)
 { int x, plen,qlen;
   if ( (x = ((MFRAG *)q)->score - ((MFRAG *)p)->score) != 0)  return x;
   /* Take the physically longer fragment: */
   /* This will tilt towards extended fragments with more mismatches */
   plen = offsetlen(((MFRAG *)p)->start, ((MFRAG *)p)->finish);
   qlen = offsetlen(((MFRAG *)q)->start, ((MFRAG *)q)->finish);
   if ( (x = qlen-plen) != 0)  return x;
   /* Take the one with the fewer mismatches */
   /* This is equivalent to choosing the one with the larger */
   /*   polymorphism length. */
   if ( (x = ((MFRAG *)p)->nmismat - ((MFRAG *)q)->nmismat) != 0)  return x;
   /* If all else fails, sort from left to right */
   if ( (x = ((MFRAG *)p)->start - ((MFRAG *)q)->start) != 0)  return x;
   return 0; }


/* Version 1.02 and previous sort: */
/* Within an overlapping tie group, this picked the fragment */
/*   whose start and finish are leftmost.

int oldsortpfrag (const void *p, const void *q)
 { int x;
   if ( (x = ((MFRAG *)q)->score - ((MFRAG *)p)->score) != 0)  return x;
   if ( (x = ((MFRAG *)p)->start - ((MFRAG *)q)->start) != 0)  return x;
   if ( (x = ((MFRAG *)p)->finish - ((MFRAG *)q)->finish) != 0)  return x;
   return 0; }

*/

static int sortstart (const void *p, const void *q)
 { int x;
   if ( (x = ((MFRAG *)q)->start - ((MFRAG *)p)->start) != 0)  return x;
   if ( (x = ((MFRAG *)q)->score - ((MFRAG *)p)->score) != 0)  return x;
   if ( (x = ((MFRAG *)p)->finish - ((MFRAG *)q)->finish) != 0)  return x;
   return 0; }



/* (***BBB) Utilities */

/* Initialize data arrays: */

static void init_cf (SIMX *sc, SIMTYPE simtype, char typechar,
   void(*do_mod)(PAIRCOMP *pc), int dosims, int npairs)
 { /* If sc->pairpc was allocated, free it. (Now sc->pairpc==NULL.) */
   if (sc->pairpc!=NULL)
    { PAIRCOMP *pc, *pcend=sc->pairpc+sc->npairs;
      for (pc=sc->pairpc; pc<pcend; pc++)
       { if (pc->kaerrmess)  free(pc->kaerrmess);
         if (pc->mf)  free(pc->mf);
         #ifdef VTEXT
         if (pc->htable)  free(pc->htable);
         if (pc->pwcnt)  free(pc->pwcnt);
         #endif
           }
      free(sc->pairpc); }
   #ifdef VTEXT
   /* If sc->gflist or sc->gdcnt were used, free them */
   if (sc->gflist)  free(sc->gflist);
   if (sc->globcnt)  free(sc->globcnt);
   #endif
   /* Clear and initialize the SIMX structure */
   memset(sc,0,sizeof(SIMX));
   sc->simtype=simtype;  sc->typechar=typechar;
   sc->do_mod=do_mod;  sc->npairs=npairs;
   sc->dosims=(dosims && numsims>0);
            }



#ifdef VTEXT
extern int ngkmax;
extern GFRAG *gfraglist;
#else /* if VSHOW */
extern float *innerdat, *outerdat;
#endif

void init_arrays(void)
 { /* Clear fragment lists and other arrays and set parameters */
   /* Enough of `Seqfile' has been read to determine the group structure */
   /*   at this point, so that npairseqs etc. are known */
   /* Init_cf sets   cf.dosims = (``allowsims'' && numsims>0). */
   /* In particular  cf.dosims==FALSE if numsims==0 . */
   init_cf(&cfinx,
       SIM_INNER, 'I', do_inner, TRUE, npairseqs);
   init_cf(&cfoutseqx,
       SIM_OUTERSEQ, 'O', do_outerseq, outersims, noutseqs);
   init_cf(&cfoutpairx,
       SIM_OUTERPAIR, 'A', do_outerpair, outersims, npairseqs);
   init_cf(&cfoutgrpx,
       SIM_OUTERGROUP, 'G', do_outergrp, outersims, noutgrps);

   /* Initialize fragment collection lists and bounds */
   #ifdef VTEXT
   listpkeep=listglobal;
   if (listpairs && listpkeep<listperpair)  listpkeep=listperpair;
   ngkmax=npkmax=npmkmax=2*listpkeep+8;
   if (ngkmax<FRAGBAT)  ngkmax=npkmax=npmkmax=FRAGBAT;

   if (currlist==NULL)
    { currlist=cxalloc(npkmax+3,sizeof(MFRAG));
      permlist=cxalloc(npmkmax+3,sizeof(MFRAG));
      gfraglist=cxalloc(ngkmax+3,sizeof(GFRAG)); }
   #else /* If VSHOW */
 { int npmax;  static int nxbases;
   /* `listpairs' is not used in VSHOW */
   listpkeep=listperpair;
   npmax=2*listperpair+8;
   if (npmax<FRAGBAT)  npmax=FRAGBAT;
   if (npmax>npkmax)
     currlist=rexxalloc(currlist,sizeof(currlist[0]), 0,npkmax=npmax);
   if (npmax>npmkmax)
     permlist=rexxalloc(permlist,sizeof(permlist[0]), 0,npmkmax=npmax);

   /* Initialize fdat arrays */
   if (nbases!=nxbases)
    { innerdat=rexxalloc(innerdat,sizeof(float), 0,nbases+4);
      outerdat=rexxalloc(outerdat,sizeof(float), 0,nbases+4);
      nxbases=nbases; }}
   #endif
     }



/* These functions add PAIRCOMP entries to sc->pairpc */
/*   in order of group and then by sequence (pair) */

static void paircount_err (SIMTYPE simtype, int rr, int npairs)
 { ErrHandler ("\nINTERNAL ERROR in init_arrays()!\n"
   "  (nseqs,npairseqs,noutseqs,noutgrps)=(%d,%d,%d,%d),\n"
   "   but (n%s)=(%d) (claimed %d)\n",
     nseqs,npairseqs, noutseqs,noutgrps, fragname(simtype),rr,npairs); }


static void init_pset (PAIRCOMP *pc, int i, int j, int g)
 { pc->seq1=i;  pc->seq2=j;  pc->group=g; }


static void init_iopairs(SIMX *sc)
 { int i,j,g,r,ra,ns, npairs=sc->npairs;  PAIRCOMP *pc;
   pc=sc->pairpc=rexxalloc(sc->pairpc,sizeof(PAIRCOMP), 0,npairs);
   for (ra=r=g=0; g<ngroups; g++, r+=ns)
    { ns=grouplist[g].ns;
      if (grouplist[g].dostat)
       { for (i=0; i<ns; i++)  for (j=i+1; j<ns; j++)
          { init_pset(pc++, r+i,r+j,g);  ra++; }}}
   sc->inmodel=TRUE;
   pc->seq1=pc->seq2=pc->group=-1;   /* To be safe */
   if (ra!=npairs)  paircount_err(sc->simtype,ra,npairs); }


static void init_oseqs(SIMX *sc)
 { int i,g,r,ra,ns, npairs=sc->npairs;  PAIRCOMP *pc;
   pc=sc->pairpc=rexxalloc(sc->pairpc,sizeof(PAIRCOMP), 0,npairs);
   for (ra=r=g=0; g<ngroups; g++, r+=ns)
    { ns=grouplist[g].ns;
      if (grouplist[g].dostat)
       { if (ns==2)  { init_pset(pc++, r,-1,g);  ra++; }
         else for (i=0; i<ns; i++)
          { init_pset(pc++, r+i,-1,g);  ra++; }}}
   sc->inmodel=TRUE;
   pc->seq1=pc->seq2=pc->group=-1;   /* To be safe */
   if (ra!=npairs)  paircount_err(sc->simtype,ra,npairs); }


static void init_ogrps(SIMX *sc)
 { int g,r,ra,ns, npairs=sc->npairs;  PAIRCOMP *pc;
   pc=sc->pairpc=rexxalloc(sc->pairpc,sizeof(PAIRCOMP), 0,npairs);
   for (ra=r=g=0; g<ngroups; g++, r+=ns)
    { ns=grouplist[g].ns;
      if (grouplist[g].dostat)
       { init_pset(pc++,-1,-1,g);  ra++; }}
   sc->inmodel=TRUE;
   pc->seq1=pc->seq2=pc->group=-1;   /* To be safe */
   if (ra!=npairs)  paircount_err(sc->simtype,ra,npairs); }


void alloc_pair_arrays (SIMX *sc)
 { /* Initialize pairwise arrays */
   switch(sc->simtype) {
     case SIM_INNER:   case SIM_OUTERPAIR:  init_iopairs(sc);  break;
     case SIM_OUTERSEQ:    init_oseqs(sc);  break;
     case SIM_OUTERGROUP:  init_ogrps(sc);  break; }}



/* Determine what P-value barriers will be imposed */
/* Usekapw, usekaglob, and useglobal_for_pairw  are static in this file */

#ifdef VTEXT

void set_pval_barriers (SIMX *sc)
 { int usesimpw, usesimglob, vnpairs=sc->npairs;
   usekapw=usekaglob=useglobal_for_pairw=FALSE;
   /* Pairwise barriers */
   usesimpw = (havsimpairpval && sc->dosims);
   if (havkapairpval || (usesimpw==FALSE && setkapairpval==FALSE))
      usekapw=TRUE;
   /* Now global barriers */
   usesimglob = (havsimglobpval && sc->dosims);
   if (havkaglobpval || (usesimglob==FALSE && setkaglobpval==FALSE))
      usekaglob=TRUE;
   sc->usekapw=usekapw;  sc->usekaglob=usekaglob;
   /* Used for setting KA barriers for fragments stored */
   /*   in PAIRCOMP arrays. */
   /* `Useglobal_for_pairw' is set in two cases: */
   /*  (i) If a more permissive global KA restriction is given, */
   /*    relax the PAIRCOMP barrier since the global list is formed */
   /*    from pairwise lists      */
   /*  (ii) If a global KA but no pairwise KA restrictions are given, */
   /*    use the global restriction even on pairwise lists. */
   if (usekaglob &&
     (usekapw==FALSE || maxkapairpval<maxkaglobpval/vnpairs))
         useglobal_for_pairw=TRUE; }


#else /* VSHOW */

void set_pval_barriers (SIMX *sc)
 { int vnpairs=sc->npairs;
   usekapw=sc->usekapw=havkapairpval;
   usekaglob=sc->usekaglob=havkaglobpval;  useglobal_for_pairw=FALSE;
   /* Pick the lower or only P-value barrier if VSHOW, */
   /*   since there is only one list */
   if (usekapw || usekaglob)
    { if (usekaglob &&
        (usekapw==FALSE || maxkaglobpval/vnpairs<maxkapairpval))
            useglobal_for_pairw=TRUE; }}


#endif



/* (***CCC) Utilites */

/* General utilities: */

char *fragname(SIMTYPE simtype)
 { switch(simtype) {
     default:  return "?????";
     case SIM_NONE:  return "no";
     case SIM_INNER:  return "inner";
     case SIM_OUTERSEQ:   return "outer-sequence";
     case SIM_OUTERPAIR:  return "outer-pair";
     case SIM_OUTERGROUP:  return "outer-group"; }}

char *Fragname(SIMTYPE simtype)
 { switch(simtype) {
     default:  return "?????";
     case SIM_NONE:  return "No";
     case SIM_INNER:  return "Inner";
     case SIM_OUTERSEQ:  return "Outer-sequence";
     case SIM_OUTERPAIR:  return "Outer-pair";
     case SIM_OUTERGROUP:  return "Outer-group"; }}


SIMX *fragsimx(SIMTYPE simtype)
 { switch(simtype) {
     default:  return NULL;
     case SIM_INNER:  return &cfinx;
     case SIM_OUTERSEQ:  return &cfoutseqx;
     case SIM_OUTERPAIR:  return &cfoutpairx;
     case SIM_OUTERGROUP:  return &cfoutgrpx; }}


SIMTYPE get_outtype(void)
 { SIMTYPE stype;
   if (outerseqfrags) stype = SIM_OUTERSEQ;
   else if (outerpairfrags) stype = SIM_OUTERPAIR;
   else if (outergrpfrags) stype = SIM_OUTERGROUP;
   else  stype=SIM_NONE;
   return stype; }


char *fragseqs(char *qbf, int i1, int i2, int g, SIMTYPE simtype)
 { switch(simtype) {
     default:  return "?????";
     case SIM_INNER:  case SIM_OUTERPAIR:
       sprintf(qbf, "%s v %s", strtag[i1], strtag[i2]);
       break;
     case SIM_OUTERSEQ:
       sprintf(qbf, "%s", strtag[i1]);  break;
     case SIM_OUTERGROUP:
       sprintf(qbf, "%s", grouplist[g].gname);  break; }
   return qbf; }


char *fragpseqs(char *qbf, PAIRCOMP *pc, SIMTYPE simtype)
 { switch(simtype) {
     default:  return "?????";
     case SIM_INNER:
       sprintf(qbf, "%s v %s (inner frags)",
          strtag[pc->seq1], strtag[pc->seq2]);  break;
     case SIM_OUTERSEQ:
       sprintf(qbf, "%s (outer sequence frags)", strtag[pc->seq1]);  break;
     case SIM_OUTERPAIR:
       sprintf(qbf, "%s v %s (outer pair frags)",
          strtag[pc->seq1], strtag[pc->seq2]);  break;
     case SIM_OUTERGROUP:
       sprintf(qbf, "%s (outer group frags)", grouplist[pc->group].gname);
       break; }
   return qbf; }


char *fragcseqs(char *q1b, PAIRCOMP *pc, SIMTYPE simtype)
 { switch(simtype) {
     default:  return "?????";
     case SIM_INNER:
       sprintf(q1b, "%s v %s, inner",
          strtag[pc->seq1], strtag[pc->seq2]);  break;
     case SIM_OUTERSEQ:
       sprintf(q1b, "%s, outer", strtag[pc->seq1]);  break;
     case SIM_OUTERPAIR:
       sprintf(q1b, "%s v %s, outer",
          strtag[pc->seq1], strtag[pc->seq2]);  break;
     case SIM_OUTERGROUP:
       sprintf(q1b, "%s, group", grouplist[pc->group].gname);
       break; }
   return q1b; }


/* Package for Karlin-Altschul approximate pairwise P-values */

/* Compute the parameters  ll,kmax
     for approximate extreme-value (Karlin-Altschul) P-values
   References:

 Karlin, S., and S. F. Altschul (1990) Methods for assessing the
statistical significance of molecular sequence features by using general
scoring schemes. PNAS87, 2264--2268.

 Karlin, S., and A. Dembo (1992) Limit distributions of maximal segmental
scores among Markov-dependent partial sums. Adv.Appl.Prob 24, 113--140.

      */


double getkascore (int score, double ll, double kmax)
 { return  ll*score - log(kmax*npolys); }

double getkapval (int score, double ll, double kmax)
 { return  1-exp(-kmax*npolys*exp(-(ll*score))); }


/* Note kapval = 1 - exp(-exp(-kascore))  */
/*             = exp(-kascore) + O[exp(-2*kascore)]  */
/*                                                   */
/* Thus -log10kapval = A kascore + O(exp(-kascore))     */
/*        where  A=log10(e)=0.434295   */


#define KACROSS  25.0

#define GCM_LOG10E    0.434294481903251827651

double getlogkapval (int score, double ll, double kmax)
 { double kasc;
   return ( (kasc=getkascore(score,ll,kmax)) < KACROSS) ?
     -log10(1.0-exp(-exp(-kasc))) : GCM_LOG10E * kasc; }


double get_kapval (double kascore)
 { return (kascore < KACROSS) ?
     1.0-exp(-exp(-kascore)) : exp(-kascore); }


double get_logkapval (double kascore)
 { return (kascore < KACROSS) ?
     -log10(1.0-exp(-exp(-kascore))) : GCM_LOG10E * kascore; }


double get_kascore (double kapval)
 { return (kapval < GCM_LOG10E * KACROSS) ?
     -log(-log(1-exp(-kapval/GCM_LOG10E))) : kapval/GCM_LOG10E; }


/* Assume that `control' ends with `\n' */

static void kaerrsave(PAIRCOMP *pc, int interr, const char *control, ...)
 { char nbuff[300];
   va_list list;  va_start(list, control);
   vsprintf(nbuff, control, list);
   if (pc->kaerrmess!=NULL)  free(pc->kaerrmess);
   pc->kaerrmess=strxdup(nbuff);  pc->kainterr=interr; }


static void kanormerr (PAIRCOMP *pc, int nd, int nn)
 { char *str;
   switch(runsimtype) {
     default:  str="?????";  break;
     case SIM_INNER:
       if (nd==0)  str="No site differences.";
       else if (nd==nn)  str="All poly sites are different.";  break;
     case SIM_OUTERSEQ:
       if (nd==0)  str="All poly sites are unique.";
       else if (nd==nn)  str="No sites are unique.";  break;
     case SIM_OUTERPAIR:
       if (nd==0)  str="All poly sites are different.";
       else if (nd==nn)  str="No site differences.";  break;
     case SIM_OUTERGROUP:
       if (nd==0)  str="All poly sites are within-group polymorphic.";
       else if (nd==nn)  str="All sites are within-group monomorphic.";
       break; }
   kaerrsave(pc, FALSE, str); }



static void showx_ka(PRINTFUNP prfun,
     SIMX *sc, int *hd, int *fout, int *onefr)
 { if (sc->inmodel)
    { char *sq="", q1b[MAXNAMETWO];
      PAIRCOMP *pc,*pcend;  int npairs=sc->npairs;
      pc=sc->pairpc;  pcend=pc+npairs;
      switch (seqtype) {
        case SILENT:  sq=" (SILENT SITES)";  break;
        case ALL:     sq=" (NUCLEOTIDES)";  break;
        case PROTEIN: sq=" (proteins)";  break; }
      for ( ; pc<pcend; pc++)  if (pc->kaerrmess)
       { if (*hd)
          { prfun("\n");
            prfun("Comparisons without any fragments%s:\n",sq);
            if (sc->simtype==SIM_INNER) *fout=TRUE;
            *hd=FALSE; }
         if (pc->totdifs==0) *onefr=TRUE;
         if (sc->simtype!=SIM_INNER && *fout)
          { prfun("\n");
            prfun("The following are for outer fragments%s:\n",sq);
            *fout=FALSE; }
         prfun("  %s: %s\n", fragpseqs(q1b,pc,sc->simtype),
            pc->kaerrmess); }}}


void show_kaparm_errs(PRINTFUNP prfun)
 { int hd=TRUE, fout=FALSE, onefr=FALSE;
   showx_ka(prfun, &cfinx, &hd, &fout, &onefr);
   showx_ka(prfun, &cfoutpairx, &hd, &fout, &onefr);
   showx_ka(prfun, &cfoutseqx, &hd, &fout, &onefr);
   showx_ka(prfun, &cfoutgrpx, &hd, &fout, &onefr);
   if (onefr)
     prfun("(The entire sequence is NOT counted as a fragment.)\n");  }



/* Timeout after KMAX iterations */

/* LKSAVE recall previously computed ll,kmax values */
/* LKBAT is the batch size for increasing a LKSAVE array */

/* See below for comments and references */

#define KMAX 3000
#define LKBAT 100

typedef struct { int mm,ndif,nn;
                 double ll,kmax; }  LKSAVE;


static void getlakmax (PAIRCOMP *pc)
 { double p,q,mp, ll0,z,y,zdel, ll,kmax;
   int i,k, mm, ndif=pc->totdifs, nn=npolys;
   static LKSAVE *lka;  static int lln,llxn;
   /* The penalty in positive form */
   mm=-pc->mismatpen;
   /* Default values: */
   pc->kavalid=FALSE;  pc->ll=0.0;  pc->kmax=1.0;
   if (ndif==0 || ndif==nn)
    { kanormerr(pc, ndif,nn);  goto showmx_out; }
   else if (nn<=0 || ndif>nn || mm<1 || (mm+1)*ndif <= nn)
    { kaerrsave (pc, TRUE,
       "LaKmax error:  totdifs=%d  nn=%d  penalty=%d  (Num=%d)",
            ndif,nn,mm, nn-(mm+1)*ndif);  goto showmx_out; }

   p=(double)ndif/nn;  q=1-p;  mp=mm*p;
   /* If the mismatch penalty is infinite: */
   if (gscale==0)
    { pc->ll=-log(q);  pc->kmax=p;  pc->kavalid=TRUE;
      goto showmx_out; }

   /* ll and kmax depend only on mm, ndif, and nn: */
   /*   Have we seen these values before? */
   for (i=0; i<lln; i++)
     if (lka[i].mm==mm && lka[i].ndif==ndif && lka[i].nn==nn)
       { pc->ll=lka[i].ll;  pc->kmax=lka[i].kmax;
         pc->kavalid=TRUE;  goto showmx_out; }

   /* Use the Newton-Raphson method to compute ll by solving   */
   /*                                                   */
   /*    phi(ll) = q*exp(ll)+p*exp(-mm*ll) = 1,  ll>0   */
   /*                                                   */
   /* Note  phi(0)=1 and phi'(0) = 1-(mm+1)*p < 0,      */
   /*    which should guarantees a solution  ll>0       */
   /*                                                   */
   /* The global minimum of phi(ll) is at ll0=log(mp/q)/(mm+1)   */

   k=0;  ll0=log(mp/q)/(mm+1);  z=exp(2*ll0);
   do { double zm;
        if (++k>=KMAX)   /* Timed out? */
         { kaerrsave (pc, TRUE,
            "LaKmax iteration timeout!  mm=%d totdifs=%d nn=%d", mm,ndif,nn);
           textprintf ("  zdel=%g  y=%g  z=%g  numiter=%d\n", zdel,y,z,k);
           goto showmx_out; }
        zm=pow(z,-mm);  z -= (zdel=(y=q*z + p*zm - 1)/(q - mp*zm/z)); }
   while (fabs(zdel)>1e-6 || fabs(y)>1e-6);

   pc->ll=ll=log(z);
   pc->kmax=kmax=(exp(ll)-1)*(q-mp*exp(-(mm+1)*ll));
   pc->kavalid=TRUE;

   /* Add these values to the queue */

 { LKSAVE aa;  memset(&aa,0,sizeof(aa));
   aa.mm=mm; aa.ndif=ndif; aa.nn=nn; aa.ll=ll; aa.kmax=kmax;
   if (lln>=llxn)
      lka=rexxalloc(lka, sizeof(LKSAVE), lln, (llxn+=LKBAT)+2);
   lka[lln++]=aa; }

showmx_out:

   /* Low-level information if -Test=<Pairlist> is entered. */
   if (match_showparms(pc->seq1,pc->seq2))
    { char q1b[MAXNAMETWO];
      int len=BPRINTF ("BLAST-like consts (%s):",
         fragcseqs(q1b,pc,runsimtype));  if (len>55) BPRINTF("\n  ");
      if (pc->kavalid)
        BPRINTF ("  lambda=%g  kmax=%g\n", pc->ll,pc->kmax);
      else  BPRINTF ("  Parameters are invalid.\n");
      BPRINTF ("  (totseqdifs=%d  npolys=%d  mismatch penalty=%d)\n",
         ndif,nn,mm); }}


/* Notes about kmax:

     The extreme-value probability is a function of
        score - log(nn*kmax)/ll = score - log(nn)/ll - log(kmax)/ll
     where

        log kmax = -2 Sum (1/k) Exp(e^{ll S_k} I_{S_k<0} + I_{S_k>=0})/k
               - log ( phi'(ll)(1-exp(ll)) )

    for random walk Sk = (random score)+(random score)+.....,

    The terms in the series are bounded by
           E(exp(ll0 S_k)) = phi(ll0)^k < 1

    If Tk = Sk + mm*k, then
       Tk+1 = Tk         with prob p   (mismatch)
            = Tk + mm+1  with prob q   (match)
    Thus
       Tk = (mm+1) * B(k,q)   and
       Sk = (mm+1) * B(k,q) - mm*k
    for binomial probabilities  B(k,q)

    HOWEVER:  Since the random walk Sk is right-continuous ---
       when it increases, it only increases by one ---
       we have an easier formula

    Kmax = (exp(ll)-1)*(q-mp*exp(-(mm+1)*ll))

    that we can use instead
       (Karlin and Dembo Adv.Appl.Prob.)
                           */


