
/* Find the most significant fragments, compute scores and statistics,
      random shuffling, etc.

   S. Sawyer - Department of Mathematics
               Washington University in St.Louis
               St. Louis, MO 63130
               email: sawyer@math.wustl.edu
               Office: 314-935-6703  FAX: 314-863-1302

   Parts of source file (line numbers approximate):

     (***AAA) (L72)  Domodel(), update global fragments

     (***BBB) (L279)  Helper functions for do_inner(),
                                do_outerseq(), etc.

     (***CCC) (L413)  Various utilites:
                              shuffle polys, etc.


   Key steps:
     Initializations:
     Find and store observed values for
       global maximum score (for all pairs of sequences)
       global frag list:  the `listglobal' largest
       pairwise frag lists:  the `listperpair' largest for each pair

   Arrays are initialized (allocated or cleared) in init_arrays()

   if (setupflag==TRUE) (the first pass for `observed' data),
     Global maximum score is stored   in set_gfrags()
     Global and pairwise observed fragments are sorted and stored
        in do_inner(), do_outerseq(), etc
        in  addfrag() and in set_gfrags()
     Pairwise fragments are copied to their arrays
        in  set_pfrags()  at the end of  do_inner(), do_outerseq(), etc
     Global fragments are copied to their arrays  in set_gfrags()

   if (setupflag==FALSE),
     Pairwise P-values for pairwise fragments
       are updated in  update_pscores()  in do_inner(), do_outerseq(), etc
     Global P-values (overall and for global fragments)
       are updated  in update_gscores()

   HOW SIMULATIONS ARE HANDLED:
     PAIRWISE:
     For each pair:
       Each simulation run generates a simulated maximum score, `simscore'
       Each PAIRCOMP structure has a pair of arrays
          int *htable;  int hlen,hxlen, ntotg;
          PWCNT *pwcnt;  int npwclen,nxpwclen;
       If `simscore'<hlen, its value is tabled in `htable'.
          The function `convert_htable', called from vtds.c,
          later converts these tabled counts in-place to P-values.
       If `simscore'>=hlen, a counter (ntotg) is incremented for the
          benefit of the tabled values.
          The observed fragment scores for that pair that are larger
          than `hlen' are stored in `pwcnt'. The counts for any of these
          fragment scores that are <= `simscore' are incremented.
       These arrays are stored in the PAIRCOMP structure and are
          independent of the observed pairwise fragment list. However,
          each MFRAG fragment structure has a pointer `grc' to one of the
          counts in the PAIRCOMP structure.
     GLOBAL:
     Bcsims==FALSE:
       Each simulation run generates a simulated maximum global score,
          `bestglobscore'
       Each SIMX structure has an array
          GLOBCNT *globcnt;  int ngloblen,nxgloblen;
       The observed score of all global fragments are sorted. Each
          distinct value is given its own GLOBCNT structure.
          The counts in these structures are incremented whenever
          `bestglobscore'>=(GLOBCNT.kascore).
          This array is stored in the SIMX structure independently of the
          observed global fragment list. However, each GFRAG global
          fragment structure contains an MFRAG structure whose pointer
          `grc' points to one of the counts in the GLOBCNT array.
          (The `grc' pointers are initialized in `set_pfrags()' before
          being copied to PAIRCOMP and SIMX fragment arrays. However, the
          `grc' pointers in the SIMX arrays are overwritten when the
          global fragment arrays are initialized in `set_gfrags()'.)
     Bcsims==TRUE:
       When the `permlist' fragment array is initialized, `set_pfrags()'
         makes sure that the `grc' pointers are initialized for all
         fragments before they are copied to SIMX fragment arrays.
       These `grc' pointers are NOT re-initialized in set_gfrags(), and
         are NOT incremented in update_gscores().
       In `trunc_glob()' in vtds.c, the values pointed to by the
         `grc' pointers in the SIMX fragment array are multiplied by
         `npairs' and copied to a new array. The `grc' pointers are then
         reassigned to the new, multiplied values.

   */



#include "vall.h"
#include "vflags.h"
#include "vread.h"
#include "vcalc.h"

char *AllRightsReserved_Vtcalc =
   "All rights reserved - Stanley Sawyer";

#define MAXHLEN 200       /* Table at most this many scores */

extern int first_offset, last_offset;

/* Prototypes for local functions */

static void set_gfrags(SIMX *sc);
static void update_gscores(SIMX *sc);


/* (***AAA) */
/* Main subfunctions, global fragments */

static int nglob, setbestglscore;
static double bestglobscore;
/* Global so that they can be initialized in vcalc.c */
GFRAG *gfraglist;
int ngkmax;


static void domodel(SIMX *sc)
 { PAIRCOMP *pc, *pcend;
   void(*do_mod)(PAIRCOMP *) = sc->do_mod;
   if (setupflag)
    { alloc_pair_arrays(sc);  set_pval_barriers(sc);
      /* Reset the global fragment collection list */
      nglob=0;  memset(gfraglist, 0, (ngkmax+3)*sizeof(GFRAG)); }
   setbestglscore=FALSE;  bestglobscore=0.0;
   runsimtype=sc->simtype;  vnumpairs=sc->npairs;
   /* Run through each pair structure, looking at/for fragments */
   pc=sc->pairpc;  pcend=pc+vnumpairs;
   while (pc<pcend)  do_mod(pc++);
   /* Copies global fragment array to *sc if setupflag==TRUE */
   if (setupflag)  set_gfrags(sc);
   else  update_gscores(sc); }



void score_data(int insetup)
 { /* Save the value of `setupflag' in a global variable */
   /* If setupflag==TRUE, clear SIMX arrays and initialize */
   /*   fragment collection lists. */
   setupflag=insetup;
   if (setupflag)
    { /* This deallocates arrays, but initializes SIMXs */
      init_arrays();
      #ifndef UNIX
      timer(TIME_INIT,0,NULL);
      #endif
       }

   /* Call do_inner(), do_outerseq(), etc. to either */
   /*   build fragment arrays and calculate scores */
   /*   or else make comparisons from simulated scores */
   if (innerpairfrags)  domodel(&cfinx);
   if (outersims || setupflag)
    { if (outerseqfrags)  domodel(&cfoutseqx);
      if (outerpairfrags)  domodel(&cfoutpairx);
      if (outergrpfrags)  domodel(&cfoutgrpx); }
   runsimtype=SIM_NONE;

   if (setupflag)
    { show_kaparm_errs(bprintf);
      #ifndef UNIX
      timer(TIME_ELAPSED,0," for building arrays");
      #endif
        }}



/* Called only if setupflag==TRUE at the end of the first pass */
/*   for all pairs */
/* Copy gfraglist[] to sc and set sc->ngfrags */

static void set_gfrags(SIMX *sc)
 { int nfrags=sc->maxnglob=nglob;   GFRAG *gf;
   sc->obsgs=bestglobscore;  sc->obvalid=setbestglscore;

   /* Return if no global fragments have been collected */
   if (nglob<=0)
    { sc->ngfrags=sc->ngloblen=0;  return; }

   /* Sort and finish up gfraglist[] */
   /* Final restrictions are done in fixup_frag_structs(). */
   /* BC Sim scores are not known at this point if allbc==TRUE, */
   /*   so that we can only sort if bcsims==FALSE */
   if (bcsims==FALSE)
    { qsort(gfraglist,nglob,sizeof(GFRAG),sortgkfrag);
      if ( (nfrags=nglob)>listglobal+3)  nfrags=listglobal+3; }
   /* Copy gfraglist[] to sc->gflist */
   /* Overlap control is applied only to pairwise lists. */
   /* If bcsims==TRUE, `gfraglist' is sorted by pair */
   /*   and then by score. */
   /* Rexxalloc() allows room for 3 additional structures. */
   if (nfrags > sc->ngxfrags)
     sc->gflist=rexxalloc(sc->gflist,sizeof(GFRAG),
           0,sc->ngxfrags=nfrags);
   gf=sc->gflist;  sc->ngfrags=nfrags;
   memmove(gf, gfraglist, nfrags*sizeof(GFRAG));
   memset(gf+nfrags, 0, sizeof(GFRAG));
   sc->ngloblen=0;
   if (bcsims==FALSE)
    { int k,nsco, *grcp=NULL;  GLOBCNT *globcnt;  double lasc;
      /* Set up arrays for efficiently computing */
      /*   global simulated P-values using BLAST-like scores. */
      /* First, count the number of distinct scores: */
      lasc=gf[0].kascore;
      for (nsco=k=1; k<nfrags; k++)  if (gf[k].kascore!=lasc)
       { nsco++;  lasc=gf[k].kascore; }
      /* Allocate an array to hold the scores. */
      if (nsco>=sc->nxgloblen)
        sc->globcnt=rexxalloc(sc->globcnt,sizeof(GLOBCNT),
           0,sc->nxgloblen=nsco);
      else  memset(sc->globcnt, 0, sc->nxgloblen*sizeof(GLOBCNT));
      globcnt=sc->globcnt;  sc->ngloblen=nsco;
      /* Store scores and pointers to counts */
      for (k=0; k<nfrags; k++)
       { if (k==0 || gf[k].kascore!=lasc)
          { lasc=globcnt->kascore=gf[k].kascore;
            grcp=&globcnt->grcount;  globcnt++; }
         gf[k].m.grc=grcp; }}}


/* Called only if setupflag==FALSE */

static void update_gscores(SIMX *sc)
 { if (setbestglscore)
    { if (bestglobscore >= sc->obsgs)  sc->nabove++;
      if (bestglobscore > sc->bestgs)  sc->bestgs = bestglobscore;
      sc->sum1 += bestglobscore;
      sc->sum2 += bestglobscore*bestglobscore;
      /* The following assumes that the scores are sorted */
      if (bcsims==FALSE)
       { int k;  GLOBCNT *gg;
         if ( (k=sc->ngloblen)>0)
           for (gg=sc->globcnt; --k>=0; )
            { if (bestglobscore >= gg[k].kascore)  gg[k].grcount++;
              else break; }}}}



int sortgkfrag (const void *p, const void *q)
 { int x;  double g;
   if ( (g = ((GFRAG *)q)->kascore - ((GFRAG *)p)->kascore) != 0.0)
        return (g>0.0) ? 1 : -1;
   if ( (x=sortpfrag(&(((GFRAG *)p)->m), &(((GFRAG *)q)->m)))!=0)
        return x;
   if ( (x = ((GFRAG *)p)->m.start - ((GFRAG *)q)->m.start) != 0)  return x;
   if ( (x = ((GFRAG *)p)->m.finish - ((GFRAG *)q)->m.finish) != 0)  return x;
   if ( (x = ((GFRAG *)p)->seq1 - ((GFRAG *)q)->seq1) != 0)  return x;
   if ( (x = ((GFRAG *)p)->seq2 - ((GFRAG *)q)->seq2) != 0)  return x;
   if ( (x = ((GFRAG *)p)->group - ((GFRAG *)q)->group) != 0)  return x;
   if ( (x = ((GFRAG *)q)->m.score - ((GFRAG *)p)->m.score) != 0)  return x;
   return 0; }


int sortgsfrag (const void *p, const void *q)
 { int x;
   if ( (x = *(((GFRAG *)p)->m.grc) - *(((GFRAG *)q)->m.grc)) != 0)
      return (x>0) ? 1 : -1;
   return sortgkfrag(p,q); }


static int sortgseq1(int s1, int s2)
 { int x;
   if (s1!=s2 && s1>=0 && s2>=0 && s1<nseqs && s2<nseqs
     && (x=stricmp(readtag[s1],readtag[s2]))!=0)
       return (x>0) ? 1 : -1;
   else return 0; }


int sortgfbyseqs (const void *p, const void *q)
 { int x, sp1,sp2, sq1,sq2;
   if ( (x = ((GFRAG *)p)->group - ((GFRAG *)q)->group) != 0)
     return (x>0) ? 1 : -1;
   sp1=((GFRAG *)p)->seq1;   sq1=((GFRAG *)q)->seq1;
   if ( (x=sortgseq1(sp1,sq1))!=0)  return (x>0) ? 1 : -1;
   sp2=((GFRAG *)p)->seq2;   sq2=((GFRAG *)q)->seq2;
   if ( (x=sortgseq1(sp2,sq2))!=0)  return (x>0) ? 1 : -1;
   return (bcsims)
     ? sortgsfrag(p,q) : sortgkfrag(p,q) ; }


/* Called after complete run of matches during simulations */

/* simcurr is the highest score for fragments */
/*   ENDING AT the current position */
/* simscore is the highest score for fragments */
/*   ENDING AT OR BEFORE the current position */
/* simfragstart is the beginning polymorphism # of a */
/*   current fragment whose score is sumcurr */

/* simcurr and simscore start initially at 0, */
/*   and are >0 iff addsim() has been called before */
/*   it was last initialized */


void addsim(int rfin)
 { static int simfragstart;
   if (circ2flag)
    { rfin+=npolys;
      /* Don't consider as fragment OR addon if length > npolys */
      /* (In that case, pos.start >= maxfragstart) */
      /* The condition simscore>0 guarantees that simfragstart is valid */
      if (simscore>0 && rfin - simfragstart > npolys)
       { adsimdone=TRUE;  return; }}
   /* Should this run be used as a fragment? */
   /* Minlength, minnpoly, and minscore are for DISPLAYED fragments */
   /*   only, and do not enter here */
   if (gscale==0 || simcurr<=0
        || pos.score > (simcurr+=(neg.score+pos.score)))
    { if (circular)
       { if (pos.start >= maxfragstart)
          { adsimdone=TRUE;  return; }
         simfragstart=pos.start; }
      simcurr=pos.score; }
   if (simcurr>simscore)  simscore=simcurr; }


/* Update the global score */
/* Called in set_pfrags() and update_pscores() */
/*   at the end of do_inner(), do_outerseq(), etc */

static void update_bestkascore(int simscore, PAIRCOMP *pc)
 { if (pc->kavalid)
    { double newscore=getkascore(simscore,pc->ll,pc->kmax);
      if (setbestglscore==FALSE || newscore>bestglobscore)
       { bestglobscore=newscore;  setbestglscore=TRUE; }}}




/* (***BBB) */
/* Pairwise functions: Helper functions for */
/*    do_inner(), do_outerseq(), etc. */


/* If setupflag==TRUE, */
/*   called at the end of do_inner(), do_outerseq(), etc for ONE pc */
/* Copy permlist[] to pc and set pc->nfrags */
/* Copy permlist[] into gfraglist[], if appropriate */

void set_pfrags (PAIRCOMP *pc)
 { int i,j,k, plen,p2len, setglgate;
   double ll,kmax, globgate;
   /* Finish up permlist[] and simscore, if necessary */
   if (pos.valid)  { addfrag(npolys);  addsim(npolys); }
   /* Use simscore to set observed pc->mxscore and update bestglobscore */
   if (simscore>0)
    { pc->mxscore=simscore;  update_bestkascore(simscore,pc); }
   /* Return if no pairwise fragments were collected. */
   if (nperm<=0)  return;
   /* Sort and finish up permlist[], */
   /*   and copy it to pc->mf */
   qsort(permlist,nperm, sizeof(MFRAG), sortpfrag);
   /* Clear out overlaps from permlist[] */
   /* Overlaps cannot occur if gscale==0 */
   if (gscale>0 && maxoverlap>0)
    { checkoverlap(2,npolys,0);  /* Initialize checkoverlap() */
      for (k=i=0; i<nperm; i++)
        if (checkoverlap(0,permlist[i].start,permlist[i].finish))
          { if (k<i)  permlist[k]=permlist[i];  k++; }
      if (k<nperm)  nperm=k;
      memset(permlist+nperm,0,sizeof(permlist[0]));
      pc->novlap=checkoverlap(1,0,0); }

   /* Define `plen', which is the number from permlist[] */
   /*    to copy to pc->mf[] */
   /* If plen>listperpair, keep an extra fragment so that later we can */
   /*   tell if the fragment list was artificially truncated. */
   /* FINAL restriction to listperpair is done in fixup_frag_structs(). */

   plen=nperm;
   if (listpairs && listperpair>0)
    { /* Fragments that are censored here can still be used */
      /*   in gfraglist[].  */
      if (plen>listperpair+1)  plen=listperpair+1;
      /* See if a pairwise KA barrier can be applied */
      /*   and we can restrict permlist[] further. */
      if (havpwbarrier && kapwbarrier>kabarrier)
       { for (k=0; k<plen; k++)  if (permlist[k].score<kapwbarrier)  break;
         if (k<plen) plen=k; }}
   /* If listpairs=bcsims=FALSE and numsims>0, keep track of */
   /*   a few fragments for possible anecdotal BC SimPvals */
   else if (numsims>0 && bcsims==FALSE) { if (plen>3) plen=3; }
   /* If listpairs==FALSE and numsims==0, no lists at all */
   else plen=0;

   /* Define `p2len', which is the maximum number from permlist[] */
   /*    to copy to gfraglist[] */
   p2len=nperm;
   if (listglobal>0)
    { if (p2len>listglobal+1)  p2len=listglobal+1;
      if (havglob_barrier && kaglob_barrier>kabarrier)
       { for (k=0; k<p2len; k++)
           if (permlist[k].score<kaglob_barrier)  break;
         if (k<p2len) p2len=k; }}
   else p2len=0;

   /* Initialize buffers and pointers for pairwise scores */
   /*   and assign `grc' pointers for permlist[] entries up to `num' */
   /* See comments at the top of `vtcalc.c' for more detail. */
   /* All counts must be cleared */
   if (plen>0 || p2len>0)
    { int num=plen;
      if (bcsims && p2len>plen)  num=p2len;
      pc->hlen=pc->npwclen=0;   /* To be safe */
      /* Set up score tables */
      /* If bcsims==TRUE,  num=max(plen,p2len) */
      /* Only `plen' of permlist[] are copied to pc->mf */
      if (num>0)
       { int *htable, *grcp, hlen,nsco,lasc;  PWCNT *pwcnt;
         if ( (hlen=simscore+1) > MAXHLEN)  hlen=MAXHLEN;
         if (hlen>pc->hxlen)
           pc->htable=rexxalloc(pc->htable,sizeof(int), 0, pc->hxlen=hlen);
         else  memset(pc->htable, 0, pc->hxlen*sizeof(int));
         htable=pc->htable;  pc->hlen=hlen;

         /* Recall that permlist[] is sorted by score, */
         /*   so that permlist[0].score=simscore */
         /* How many odd scores > MAXHLEN in permlist[], up to num? */
         for (nsco=k=0; k<num; k++)
           if (k==0 || permlist[k].score!=lasc)
            { if ( (lasc=permlist[k].score) >= hlen)  nsco++;
              else break; }
         if (nsco>0)
          { if (nsco>pc->nxpwclen)
              pc->pwcnt=rexxalloc(pc->pwcnt,sizeof(PWCNT),
                 0, pc->nxpwclen=nsco);
            else  memset(pc->pwcnt, 0, pc->nxpwclen*sizeof(PWCNT)); }
         /* There are `nsco' distinct values > MAXHLEN */
         /*   in permlist[] up to `num' */
         pwcnt=pc->pwcnt;  pc->npwclen=nsco;

         /* Store scores and pointers for permlist[] fragments */
         /* If  permlist[k].score<hlen,  permlist[k].grc */
         /*   points to the `score'th table entry in  pc->htable */
         /* If  permlist[k].score>=hlen, permlist[k].grc */
         /*   points to grcp, which is in a PWCNT entry in  pc->pwcnt  */
         /* int *grcp and int lasc are initialized when k==0 */
         /* Note that these score structures and their number */
         /*   are independent of pc->mf and its length, although */
         /*   the `grcp' pointers are stored in permlist[] */
         for (k=0; k<num; k++)
          { if (k==0 || permlist[k].score!=lasc)
             { if ( (lasc=permlist[k].score) < hlen)
                 grcp=permlist[k].grc=&htable[lasc];
               else
                { pwcnt->score=lasc;
                  grcp=permlist[k].grc=&pwcnt->grcount;  pwcnt++; }}
            else permlist[k].grc=grcp; }}}

   /* Low-level information if -Test=<Pairlist> is entered. */
   if (match_showparms(pc->seq1,pc->seq2))
    { char q1b[MAXNAMETWO];
      BPRINTF ("FRAGMENTS COLLECTED (%s):  global=%d, pairwise=%d\n",
        fragcseqs(q1b,pc,runsimtype), p2len,plen); }

   /* Store the pairwise and global lists. */
   /* These steps MUST NOT be done before permlist[].grc pointers */
   /*   are assigned. */
   /* FINAL restriction to listperpair is in fixup_frag_structs(). */
   if (plen>0)
    { if (plen > pc->nxfrags)
          pc->mf=rexxalloc(pc->mf, sizeof(MFRAG), 0, pc->nxfrags=plen);
      memmove(pc->mf, permlist, plen*sizeof(MFRAG));
      /* Recall that rexxalloc() allows room for 3 more */
      memset((pc->mf)+plen, 0, 2*sizeof(MFRAG)); }
   pc->nfrags=plen;

   /* Finally, add fragments from permlist[] to gfraglist[] */

   if (pc->kavalid)  { ll=pc->ll;  kmax=pc->kmax; }
   else if (bcsims==FALSE)
    { char q1b[2*MAXNAMEBUFF+33];
      printf ("WARNING (%s): K-A parameters are invalid\n"
         "  (NOT INCLUDED in global fragment lists)\n",
             fragcseqs(q1b,pc,runsimtype));  return; }

   /* Note that the following copies `permlist[].grc' pointers */
   /*   to gfraglist[]. These points are used only for reading. */
   /*   If bcsims==TRUE, they are used to generate BC SimPvals. */
   /*   (The function `trunc_glob()' in vtds.c multiplies the P-vales */
   /*      by npairs by creating a new score array.) */
   /*   If bcsims==FALSE, the `permlist[].grc = gfraglist[].m.grc' */
   /*     values are overwritten in `set_gfrags()' */
   /* See comments at the top of `vtcalc.c' for more detail */
   /* Note that permlist[] for this PAIRCOMP is already sorted */
   /*   by score, and that overlaps have been weeded out. */
   /* Thus if one fragment is below globgate, then all succeeding */
   /*   fragments are below globgate as well. */
   /* NOTE: Fragment buffer expansion is NOT usually necessary here, */
   /*   since ngkmax>=2*listglobal+8.  Thus the only pairwise fragments */
   /*   that are lost are BELOW the best listglobal SO FAR.  Recall */
   /*   that overlap control is applied to pairwise lists ONLY. */
   /* This restricts to between listglobal+1 and ngkmax. */
   /*   Final restrictions are done in fixup_frag_structs(). */

   setglgate=FALSE;  globgate=0.0;
   for (j=0; j<p2len; j++)
    { double newscore=getkascore(permlist[j].score,ll,kmax);
      GFRAG g;
      if (bcsims==FALSE)
       { if (setglgate && newscore < globgate)  break;
         if (nglob>=ngkmax)  /* Flush gfraglist[] back to listglobal */
          { qsort(gfraglist,nglob, sizeof(GFRAG),sortgkfrag);
            nglob=listglobal+1;  setglgate=TRUE;
            /* Set new barrier. */
            globgate=gfraglist[listglobal-1].kascore;
            /* Is the new fragment still of global interest? */
            if (newscore<globgate)  break; }}
      /* If bcsims==TRUE, gfraglist[] CANNOT be limited */
      /*   to listglobal+1, but can be limited to (listglobal+1)*npairs */
      else if (nglob>=ngkmax)
       { if (listglobal>FRAGBAT)  ngkmax+=2*listglobal;
         else  ngkmax+=2*FRAGBAT;
         gfraglist=rexxalloc(gfraglist,sizeof(GFRAG), nglob,ngkmax); }
      memset(&g,0,sizeof(g));
      g.seq1=pc->seq1;  g.seq2=pc->seq2;  g.group=pc->group;
      g.mismatpen=pc->mismatpen;  g.totdifs=pc->totdifs;
      g.kascore=newscore;  g.m=permlist[j];
      gfraglist[nglob++]=g; }}



/* If setupflag==FALSE, */
/*   called at the end of do_inner(), do_outerseq(), etc */
/*   after simscore has been calculated */

void update_pscores(PAIRCOMP *pc)
 { /* Finish up simscore if necessary */
   if (pos.valid)  addsim(npolys);
   /* Use simscore to update bestglobscore */
   if (simscore>0)  update_bestkascore(simscore,pc);
   /* Use simscore to update pairwise fragment P-values */
   if (pc->hlen>0)
    { if (simscore<pc->hlen)  pc->htable[simscore]++;
      else
       { int k;  PWCNT *gg;  pc->ntotg++;
         /* The following assumes that the scores are sorted */
         if ( (k=pc->npwclen)>0)
           for (gg=pc->pwcnt; --k>=0; )
            { if (simscore>=gg[k].score)  gg[k].grcount++;
              else break; }}}}



/* (***CCC) */
/* Various utilities */


static void gen_ghs (SIMX *sc)
 { if (sc->inmodel && sc->dosims)
    { PAIRCOMP *pc,*pcend;  int *htable;
      pc=sc->pairpc;  pcend=pc+sc->npairs;
      for ( ; pc<pcend; pc++)  if ( (htable=pc->htable)!=NULL)
       { int hc=pc->hlen, sumb=pc->ntotg;
         while (--hc>=0)  sumb=(htable[hc]+=sumb); }}}


void convert_htable(void)
 { gen_ghs(&cfinx);
   gen_ghs(&cfoutseqx);
   gen_ghs(&cfoutpairx);
   gen_ghs(&cfoutgrpx); }




/* Randomly shuffle the  polysites[]  array */

static void shuffall(void)
 { int i = npolys, j;
   while (--i > 0)
    { char *temp = polysites[i].site;
      polysites[i].site = polysites[j = nrand(i+1)].site;
      polysites[j].site = temp; }}


static void shuffbydeg (void)
 { int i,j, n;  POLYSITE  *rr;
   static char **plist[6];  static int ndeg[6];
   if (plist[0]==NULL)
    { for (rr=polysites; rr<sitend; rr++)
       { if ( (i=rr->degen+1) >= 0 && i < 6)  ndeg[i]++;
         else ErrHandler ("\nINTERNAL ERROR: UNRECOGNIZED DEGENERACY %d\n",
                  rr->degen); }
      for (i=0; i<6; i++)
       { char **pl = plist[i] = cxalloc(ndeg[i]+3, sizeof(char **));
         if (ndeg[i]>1)
          { int deg=i-1;
            for (j=0, rr=polysites; rr<sitend; rr++)  if (rr->degen==deg)
              pl[j++] = rr->site; }}}
   for (i=0; i<6; i++)  if ( (n=ndeg[i])>1)
    { char *ptemp, **pl=plist[i];  int deg=i-1;
      for (rr=polysites; rr<sitend; rr++)  if (rr->degen==deg)
       { ptemp = pl[--n];
         rr->site = pl[n] = pl[j = nrand(n+1)];  pl[j] = ptemp; }}}


void shuffdata(void)
 { if (seqtype==SILENT && nodegenflag==FALSE)  shuffbydeg();
   else  shuffall(); }


